unit Gauge;

interface

uses NUtil;

type
  TLogger = procedure (Sender: TObject; const S: string) of object;

  TGauge = class(TObject)
  private
    Name: string;
    procedure SetValue(const Name, Value: string);
  public
    Symbols: TStrDict;
    DefaultInterval: Integer;
    UpdateTime: Integer;
    Log: TLogger;
    constructor Create(const ASymbols: TStrDict; const ALogger: TLogger);
    procedure Init; virtual;
    procedure Update; virtual;
    function Conf(const Name, Default: string): string;
    function Exists(const Name: string): Boolean;
    property Value[const Name: string]: string write SetValue; default;
  end;
  TGaugeClass = class of TGauge;

  TGauges = class(TObject)
  private
    Gauges: TObjList;
    function NameToClass(const Name: string): TGaugeClass;
    function GetValue(const Name: string): string;
    procedure SetConf(const Name, Value: string);
  public
    Symbols: TStrDict;
    Log: TLogger;
    constructor Create;
    destructor Destroy; override;
    procedure Update;
    procedure Clear;
    function EachSymbols(var Name: string): Boolean;
    property Value[const Name: string]: string read GetValue; default;
    property Conf[const Name: string]: string write SetConf;
  end;

implementation

uses Windows, NSys, NLib, Builtin;

{ TGauge }

constructor TGauge.Create(const ASymbols: TStrDict; const ALogger: TLogger);
begin
  Name := Lower(Copy(ClassName, 2, Length(ClassName)-1));
  Symbols := ASymbols;
  Log := ALogger;
  DefaultInterval := 1;
  Init;
  Log(Self, 'starting, update interval ' + IStr(DefaultInterval) + ' sec.');
end;

procedure TGauge.SetValue(const Name, Value: string);
begin
  Symbols[Self.Name + '.' + Name] := Value;
end;

function TGauge.Exists(const Name: string): Boolean;
begin
  Result := Symbols.HasKey(Self.Name + '.' + Name);
end;

function TGauge.Conf(const Name, Default: string): string;
var
  s: string;
begin
  s := 'conf.' + Self.Name + '.' + Name; 
  if not Symbols.HasKey(s) then
    Symbols[s] := Default;
  Result := Symbols[s]
end;

procedure TGauge.Init;
begin
end;

procedure TGauge.Update;
begin
end;

{ TGauges }

constructor TGauges.Create;
begin
  Symbols := TStrDict.Create(257);
  Gauges := TObjList.Create;
end;

destructor TGauges.Destroy;
begin
  Symbols.Free;
  Gauges.Free;
end;

procedure TGauges.Update;
var
  i: Integer;
  t: Integer;
begin
  for i := 0 to Gauges.Count-1 do
    with Gauges[i] as TGauge do
    begin
      t := GetTickCount;
      if (UpdateTime = 0) or
        ((t - UpdateTime) div 1000 >= DefaultInterval) then
      begin
        UpdateTime := t;
        (Gauges[i] as TGauge).Update;
      end;
    end;
end;

procedure TGauges.Clear;
begin
  Gauges.Clear;
  Symbols.Clear;
end;

function TGauges.NameToClass(const Name: string): TGaugeClass;
var
  i, j: Integer;
begin
  i := 0;
  while i < Length(Name) do
  begin
    Inc(i);
    if Name[i] = '.' then
      Break
  end;

  for j := 0 to High(GaugeClasses) do
    if GaugeClasses[j].ClassNameIs('T'+Copy(Name, 1, i-1)) then
    begin
      Result := GaugeClasses[j];
      Exit;
    end;
  Result := nil;
end;

function TGauges.GetValue(const Name: string): string;
  function GetPluginName(const Name: string): string;
  var
    i, j: Integer;
  begin
    i := 0;
    while i < Length(Name) do
    begin
      Inc(i);
      if Name[i] = '.' then
        Break
    end;
    j := i+1;
    while i < Length(Name) do
    begin
      Inc(i);
      if Name[i] = '.' then
      begin
        Result := Copy(Name, j, i-j);
        Exit;
      end;
    end;
    Result := Copy(Name, j, i-j+1);
  end;
var
  i: Integer;
  c: TGaugeClass;
begin
  if (Name <> '') and (Name[1] in ['-', '0'..'9']) then // literal
  begin
    Result := Name;
    Exit;
  end;

  if not Symbols.HasKey(Name) then // create key at first
    SetDefaultValue(Symbols, Name);

  c := NameToClass(Name);
  if Assigned(c) then
  begin
    for i := 0 to Gauges.Count-1 do
      if (Gauges[i] is TPlugin) then
      begin
        if GetPluginName(Name) = (Gauges[i] as TPlugin).PluginName then
        begin
          Result := Symbols[Name];
          Exit;
        end;
      end
      else if Gauges[i] is c then  // already created
      begin
        Result := Symbols[Name];
        Exit;
      end;

    if c = TPlugin then
      Gauges.Add(TPlugin.CreateByPlugin(Symbols, Log,
        GetPluginName(Name)))
    else
      Gauges.Add(c.Create(Symbols, Log));
  end;
  Result := Symbols[Name];
end;

function TGauges.EachSymbols(var Name: string): Boolean;
begin
  Result := Symbols.EachKeys(Name);
end;

procedure TGauges.SetConf(const Name: string; const Value: string);
begin
  Symbols['conf.'+Name] := Value;
end;

end.
