unit NHttp;

interface

uses NSys, NLib, NUtil, NSock; 

type
  THttpResponse = record
    Status: Integer;
    ReasonPhrase: string;
    Received: Boolean;
    LastModified: string;
    Body: string;
  end;

  THttp = class(TObject)
    Tcp: TNTcp;
    Host: string;
    Port: Integer;
    Path: string;
    Buf: string;
    Url: string;
    Error: string;
    Info: string;
    Closed: Boolean;
    IMS: string;
    UserAgent: string;
    Response: THttpResponse;
    Callback: procedure (Sender: TObject) of object;
    constructor Create;
    destructor Destroy; override;
    procedure Get(const AUrl: string; const AIms: string);
    procedure ParseResponse;
    procedure OnGetHostByName(const IpAddr: string);
    procedure OnConnect(Sender: TObject);
    procedure OnRecv(Sender: TObject);
    procedure OnClose(Sender: TObject);
    procedure OnError(Sender: TObject; ErrorCode: Integer);
  end;

implementation

procedure UrlSplit(const S: string; var Scheme, NetLoc, Path: string);
var
  i: Integer;
  tmp: string;
begin
  i := 0;
  Scheme := '';
  NetLoc := '';
  Path := '';
  while i < Length(S) do
  begin
    Inc(i);
    case S[i] of
      ':':
        if Scheme = '' then
        begin
          Scheme := tmp;
          tmp := '';
        end else
          tmp := tmp + S[i];
      '/':
        if (NetLoc = '') and (tmp <> '/') then
        begin
          NetLoc := tmp;
          tmp := '';
        end else
          tmp := tmp + S[i];
    else
      tmp := tmp + S[i];
    end;
  end;
  Path := '/' + tmp;
end;

constructor THttp.Create;
begin
  Tcp := TNTcp.Create;
  Tcp.OnGetHostByName := OnGetHostByName;
  Tcp.OnConnect := OnConnect;
  Tcp.OnRecv := OnRecv;
  Tcp.OnClose := OnClose;
  Tcp.OnError := OnError;
end;

destructor THttp.Destroy;
begin
  Tcp.Free;
end;

procedure THttp.Get(const AUrl: string; const AIms: string);
var
  i: Integer;
  s: string;
  scheme, loc: string;
begin
  Host := '';
  Buf := '';
  Error := '';
  Info := '';
  Closed := False;
  Url := AUrl;
  IMS := AIms;
  Response.Received := False;
  UrlSplit(Url, scheme, loc, Path);
  for i := 1 to Length(loc) do
    if (loc[i] = ':') and (Host = '') then
    begin
      Host := s;
      s := '';
    end else
      s := s + loc[i];

  if Host = '' then
  begin
    Host := loc;
    Port := 80;
  end else
    Port := SInt(s);

  Tcp.GetHostByName(Host);
end;

procedure THttp.ParseResponse;
  procedure ParseStatusLine(const S: string; var Code: Integer;
    var Phrase: string);
  var
    i, len: Integer;
  begin
    Code := 0;
    len := Length(S);
    if len > 0 then
    begin
      if S[len] = #10 then
        Dec(len);
      if S[len] = #13 then
        Dec(len);
    end;
    for i := 1 to len do
      if S[i] = ' ' then
      begin
        Code := SInt(Copy(S, i+1, 3));
        Phrase := Copy(S, i+5, len-(i+5)+1);
        Exit;
      end;
  end;
  procedure ParseHeader(const S: string; var Name, Value: string);
  var
    i, len: Integer;
  begin
    len := Length(S);
    if len > 0 then
    begin
      if S[len] = #10 then
        Dec(len);
      if S[len] = #13 then
        Dec(len);
    end;
    for i := 1 to len do
      if S[i] = ':' then
      begin
        Name := Copy(S, 1, i-1);
        Value := Copy(S, i+2, len-(i+2)+1);
        Exit;
      end;
    Name := '';
    Value := '';
  end;
var
  eoh: Boolean;
  s, n, v: string;
begin
  Response.Status := 0;
  Response.Body := '';
  Response.LastModified := '';
  eoh := False;
  with TSplitLines.Create(Buf) do
  begin
    while GetLine(s) do
      if Response.Status = 0 then
        ParseStatusLine(s, Response.Status, Response.ReasonPhrase)
      else if eoh then
        Response.Body := Response.Body + s
      else
        if s = #13#10 then
          eoh := True
        else begin
          ParseHeader(s, n, v);
          if Lower(n) = 'last-modified' then
            Response.LastModified := v;
        end;
    Free;
  end;
  Response.Received :=  Response.Status > 0;
end;

procedure THttp.OnGetHostByName(const IpAddr: string);
begin
  if IpAddr = '' then
  begin
    Error := 'Host not found';
    Info := '';
    Closed := True;
    if Assigned(Callback) then
      Callback(Self);
    Exit;
  end;

  Info := Fmt('Connecting to %s...', [Host]);
  if Assigned(Callback) then
    Callback(Self);

  Tcp.Connect(IpAddr, Port);
end;

procedure THttp.OnConnect(Sender: TObject);
var
  header: string;
begin
  Closed := False;
  Info := Fmt('Connected. Sending request to %s...', [Host]);
  if Assigned(Callback) then
    Callback(Self);
  header := Fmt('Host: %s'#13#10, [Host]);
  if IMS <> '' then
    header := header + Fmt('If-Modified-Since: %s'#13#10, [IMS]);
  if UserAgent <> '' then
    header := header + Fmt('User-Agent: %s'#13#10, [UserAgent]);
  (Sender as TNTcp).Send(Fmt('GET %s HTTP/1.0'#13#10'%s'#13#10,
    [Path, Header]));
end;

procedure THttp.OnRecv(Sender: TObject);
var
  s: string;
begin
  repeat
    s := (Sender as TNTcp).Recv(1024);
    Buf := Buf + s;
  until s = '';
end;

procedure THttp.OnClose(Sender: TObject);
begin
  Closed := True;
  if Buf <> '' then
  begin
    ParseResponse;
    Info := Fmt('Received response. %s %s',
      [Response.Status, Response.ReasonPhrase]);
  end;
  if Assigned(Callback) then
    Callback(Self);
end;

procedure THttp.OnError(Sender: TObject; ErrorCode: Integer);
begin
  Error := SysError(ErrorCode);
  Info := '';
  Closed := True;
  if Assigned(Callback) then
    Callback(Self);
end;

end.
